from typing import Optional
from pyrsistent import PClass, field
from pathlib import Path


from pathlib import Path
from eliot import FileDestination
import sys
from datetime import datetime
from eliottree import tasks_from_iterable, render_tasks
from eliot import Logger
from eliot.json import _dumps_bytes, _dumps_unicode, json_default

from eliot._output import *
import uuid
from enum import IntEnum
from io import TextIOBase, IOBase


import time
from functools import wraps, partial
from typing import Optional, Union, Sequence
from eliot import log_call
from enum import IntEnum

class LogLevel(IntEnum):
    """Enumeration of log levels with their corresponding numeric values"""
    NONE = 0
    DEBUG = 10
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50

def format_time(seconds: float) -> str:
    hours = int(seconds // 3600)
    seconds %= 3600
    minutes = int(seconds // 60)
    seconds = int(seconds % 60)
    return f"{hours:02}:{minutes:02}:{seconds:02}"

def log_function(
    wrapped_function=None,
    action_type=None,
    include_args=None,
    include_result=True,
    log_level: int = 20,  # INFO level by default
    include_timing: bool = False
):
    """Decorator/decorator factory that logs inputs, return result, and optionally execution time.

    @param action_type: The action type to use. If not given the function name will be used.
    @param include_args: If given, should be a list of strings, the arguments to log.
    @param include_result: True by default. If False, the return result isn't logged.
    @param log_level: Integer indicating the logging level (default: 20/INFO)
    @param include_timing: If True, logs execution time (default: False)
    """
    if wrapped_function is None:
        return partial(
            log_call,
            action_type=action_type,
            include_args=include_args,
            include_result=include_result,
            log_level=log_level,
            include_timing=include_timing
        )

    # ... existing action_type and include_args validation ...

    @wraps(wrapped_function)
    def _action(this, *args, **kwargs):
        from eliot import Message, start_action

        if include_timing:
            start_time = time.perf_counter()

        with start_action(action_type=action_type) as ctx:
            # ... existing argument logging logic ...

            result = wrapped_function(this, *args, **kwargs)

            if include_timing:
                end_time = time.perf_counter()
                execution_time = end_time - start_time
                Message.log(
                    message_type="function_timing",
                    level=log_level,
                    message=f"{wrapped_function.__name__} took {format_time(execution_time)}"
                )

            if include_result:
                ctx.add_success_fields(result=result)

            return result

    # ... rest of the existing code ...

class RenderingFileDestination(FileDestination):
    """
    A FileDestination that also renders messages in a human-readable format.
    
    Args:
        json_file: File object for JSON log output (can be Path or TextIO)
        rendered_file: File object for human-readable log output (can be Path or TextIO)
        encoder: Optional custom JSON encoder
    """
    rendered_file = field(mandatory=True)

    def __new__(cls, json_file, rendered_file, encoder=None):
        # Handle both Path/string and TextIO objects for json_file
        if isinstance(json_file, (str, Path)):
            json_file = open(json_file, 'a')
        
        # Handle both Path/string and TextIO objects for rendered_file
        if isinstance(rendered_file, (str, Path)):
            rendered_file = open(rendered_file, 'a')

        return PClass.__new__(cls,
                            file=json_file,
                            rendered_file=rendered_file,
                            _dumps=_dumps_unicode if isinstance(json_file, (IOBase, TextIOBase)) else _dumps_bytes,
                            _linebreak="\n" if isinstance(json_file, (IOBase, TextIOBase)) else b"\n",
                            _json_default=json_default)

    def __call__(self, message):
        # First let parent class handle the JSON file writing
        super().__call__(message)
        
        try:
            # Convert the message to tasks and render them
            tasks = list(tasks_from_iterable([message]))
            render_tasks(self.rendered_file.write, tasks, colorize=False, human_readable=True)
            self.rendered_file.flush()
        except Exception as e:
            self.rendered_file.write(f"Error rendering message: {str(e)}\n")
            self.rendered_file.flush()

def to_nice_stdout(output_file: Optional[Path] = None, encoder=None, json_default=json_default):
    """Configure Eliot logging with improved rendering to stdout
    
    Args:
        output_file (Optional[Path]): Path to the JSON log file. If None, creates a temporary file
        encoder: Optional custom JSON encoder
        json_default: JSON serialization function for unknown types
    """
    if output_file is None:
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        temp_dir = Path("/tmp/just_semantic_search_logs")
        temp_dir.mkdir(parents=True, exist_ok=True)
        output_file = temp_dir / f"log_{timestamp}_{uuid.uuid4().hex[:8]}.json"
    
    destination = RenderingFileDestination(
        json_file=output_file,
        rendered_file=sys.stdout
    )
    Logger._destinations.add(destination)


def to_nice_file(output_file: Path, rendered_file: Path, encoder=None, json_default=json_default):
    """Configure Eliot logging with improved rendering
    
    Args:
        output_file (Path): Path to the JSON log file
        rendered_file (Path): Path to the human-readable rendered log file
        encoder: Optional custom JSON encoder
        json_default: JSON serialization function for unknown types
    """
    destination = RenderingFileDestination(
        json_file=output_file,
        rendered_file=rendered_file
    )
    Logger._destinations.add(destination)
    # stdout: TextIO 


from pathlib import Path
from typing import Union, Optional, Callable
from functional import seq
from functional.pipeline import Sequence


def children(p: Union[Path, str]) -> Sequence:
    """
    Returns a sequence of all files and subfolders in the given directory.
    
    Args:
        p: Path or string path to the directory
    """
    path = Path(p) if isinstance(p, str) else p
    return seq(list(path.iterdir()))


def dirs(p: Union[Path, str]) -> Sequence:
    """
    Returns a sequence of only the subdirectories in the given directory.
    
    Args:
        p: Path or string path to the directory
    """
    path = Path(p) if isinstance(p, str) else p
    return children(path).filter(lambda f: f.is_dir())


def files(p: Path) -> Sequence:
    """
    Returns a sequence of only the files in the given directory.
    
    Args:
        p: Path to the directory
        
    Returns:
        Sequence of Path objects representing files
    """
    return children(p).filter(lambda f: f.is_file())


def with_ext(p: Path, ext: str) -> Sequence:
    """
    Returns files in the directory that have the specified extension.
    
    Args:
        p: Path to the directory
        ext: File extension to filter by (e.g., '.txt', '.py')
        
    Returns:
        Sequence of Path objects representing matching files
    """
    return files(p).filter(lambda f: ext in f.suffix)


def by_ext(p: Path, ext: str) -> Sequence:
    return files(p).filter(lambda f: ext in f.suffix).group_by(lambda f: f.suffix)


def rename_files_with_dictionary(files_or_path: Union[Sequence, Path, str], dictionary: dict) -> list[tuple[str, str]]:
    """
    Renames files based on a dictionary of old->new substring pairs.
    
    Args:
        files_or_path: Path/string to a directory/file or a sequence of files
        dictionary: Dictionary mapping old substrings to new substrings
    """
    if isinstance(files_or_path, str):
        files_or_path = Path(files_or_path)
    
    if isinstance(files_or_path, Path):
        if files_or_path.is_dir():
            return rename_files_with_dictionary(files(files_or_path), dictionary=dictionary)
        else:
            return rename_files_with_dictionary(seq(files_or_path),  dictionary=dictionary)
    else:
        results = []
        for p in files_or_path:
            for k, v in dictionary.items():
                if k in p.name:
                    new_name = p.name.replace(k, v)
                    p.rename(Path(p.parent, p.name.replace(k, v)))
                    results.append((p.name, new_name))
        return results


def rename_files(files_or_path: Union[Sequence, Path], has: str, what: str, to: str):
    """
    rename files that contain a substring
    :param files_or_path: sequence of files or folder #TODO: update to Union[seq, Path] when switched to python 3.10
    :param has: substring to search for
    :param what: substring to substitute (not always the same as "has")
    :param to: substitute to string
    :return: renamed files
    """
    if isinstance(files_or_path, Path):
        if files_or_path.is_dir():
            return rename_files(files(files_or_path), has, what, to)
        else:
            return rename_files(seq(files_or_path), has, what, to)
    else:
        return files_or_path.map(lambda p: p if has not in p.name else p.rename(Path(p.parent, p.name.replace(what, to))))


def rename_not_files(files: Sequence, not_has: str, what: str, to: str) -> Sequence:
    """
    rename files that do NOT contain a substring
    :param files: sequence of files
    :param not_has: substring to avoid
    :param what: substring to substitute (not always the same as "has")
    :param to: substitute to string
    :return: renamed files
    """
    return files.map(lambda p: p if not_has in p.name else p.rename(Path(p.parent, p.name.replace(what, to))))


def replace_in_file(file: Path, what: str, to: str, output: Optional[Path] = None) -> Path:
    """
    Replaces text in the file
    :param file: file to make replacement
    :param what: which text to replace
    :param to: to what string
    :param output: path to the output file (or same file if no output provided)
    :return:
    """
    in_place = output is None
    with file.open("r+") as text_file:
        s: str = text_file.read().replace(what, to)
        if in_place:
            text_file.write(s)
            return file
        else:
            output.write_text(s)
            return output


def replace_from_dict_in_file(file: Path, replacement: dict, output: Optional[Path] = None, verbose: bool = False) -> Path:
    """
    :param file: Path to the file
    :param replacement: dictionary for replacing test in files
    :param output:
    :param verbose:
    :return:
    """
    in_place = output is None
    with file.open("r+") as text_file:
        s: str = text_file.read()
    for old, new in replacement.items():
        if old in s:
            if verbose:
                print(f"REPLACING {old}\n WITH {new}")
            s = s.replace(old, new) # warning: mutation!
    if in_place:
        if verbose:
            print(f"editing {str(file)} in place")
        with file.open("w") as rewrite:
            rewrite.write(s)
        return file
    else:
        if verbose:
            print(f"writing {str(file)} with replacements to {str(output)}")
        output.write_text(s)
        return output


def traverse(p: Union[Path, str], fun: Callable[[Path], bool] = None, max_depth: int = -1, flatten: bool = True, depth: int = 0) -> list:
    """
    Recursively traverses a directory structure applying an optional filter function.
    
    Args:
        p: Path or string path to start traversal from
        fun: Optional filter function that takes a Path and returns bool
        max_depth: Maximum depth to traverse (-1 for unlimited)
        flatten: If True, returns a flat list; if False, maintains directory structure
        depth: Current traversal depth (used internally)
        
    Returns:
        List of Path objects that match the filter criteria
    """
    path = Path(p) if isinstance(p, str) else p
    fl = files(path).to_list() if fun is None else files(path).filter(fun).to_list()
    folds = dirs(path).to_list() if fun is None else dirs(path).filter(fun).to_list()
    if depth == max_depth:
        return fl + folds
    else:
        if flatten:
            return fl + folds + dirs(path).flat_map(lambda d: traverse(d, fun, max_depth, flatten, depth + 1)).to_list()
        else:
            return fl + folds + dirs(path).map(lambda d: traverse(d, fun, max_depth, flatten, depth + 1)).to_list()


def tprint(p: Path, max_depth: int = -1,  prefix: str = "", debug: bool = False, depth: int = 0):
    """
    Pretty-print the content of the folder recursively
    :param p: path to print content for
    :param max_depth: how deep to traverse, by default -1 which is unlimited
    :param prefix: prefix to add in the beginning
    :param debug: adding debug statements to separate files from folders
    :param depth: current depth
    :return:
    """
    fl = files(p)
    print(prefix+p.name)
    if fl.len() > 0:
        if debug:
            print(prefix + "FILES:")
        files(p).for_each(lambda f: print(f"\t"+prefix+f.name))
    folders = dirs(p)
    if folders.len() > 0 and depth != max_depth:
        if debug:
            print(prefix+"FOLDERS")
        folders.for_each(lambda d: tprint(d, max_depth=max_depth, prefix = f"\t"+prefix, debug=debug, depth=depth+1))


import sys
from enum import Enum
from typing import Union
import warnings

import dotenv
from dotenv import load_dotenv
import os
from deprecated import deprecated

from pycomfort.logging import LogLevel

try:
    from loguru import logger
    HAS_LOGURU = True
except ImportError:
    HAS_LOGURU = False



LOG_LEVELS = [loader.value for loader in LogLevel]

@deprecated(
    reason="Please configure loguru directly in your application. We will remove this in future versions",
    version='0.0.17'
)
def configure_logger(log_level: Union[str, LogLevel], add_stdout: bool = False):
    """Configure loguru logger with specified log level and stdout options
    
    Args:
        log_level: Either a LogLevel enum or string matching a log level
        add_stdout: If True, adds stdout as a log sink
    
    Returns:
        Configured logger instance
    
    Raises:
        ImportError: If loguru package is not installed
    """
    if not HAS_LOGURU:
        raise ImportError("loguru is required for configure_logger but it is not installed. "
                        "Install it with: pip install loguru")
    
    level = log_level.value if isinstance(log_level, LogLevel) else log_level
    if level.upper() != LogLevel.NONE.value and add_stdout:
        logger.add(sys.stdout, level=level.upper())
    return logger


def load_environment_keys(debug: bool = True, usecwd: bool = False):
    """Load OpenAI API key from .env file
    
    Args:
        debug: If True, prints debug information about env file location
        usecwd: If True, searches for .env file in current working directory
               If False, searches in parent directories
    
    Returns:
        str: OpenAI API key if found, None otherwise
    """
    # Find .env file location
    e = dotenv.find_dotenv(usecwd=usecwd)
    print(e)
    if debug:
        print(f"environment found at {e}")
    
    # Load environment variables from .env file
    has_env: bool = load_dotenv(e, verbose=True, override=True)
    if not has_env:
        print("Did not found environment file, using system OpenAI key (if exists)")
    
    # Get OpenAI key from environment variables
    openai_key = os.getenv('OPENAI_API_KEY')
    return openai_key


import json
from pathlib import Path
from typing import Optional

import typer
from pycomfort.files import replace_in_file, replace_from_dict_in_file

app = typer.Typer()

@app.command("replace")
def replace(
    file: Path = typer.Option(..., exists=True, help="rename file with substitution"),
    what: str = typer.Option(..., help="substitute --from"),
    to: str = typer.Option(..., help="substitute --to"),
    output: Optional[Path] = typer.Option(None, help="optional output, will rewrite --file is not output provided")
) -> Path:
    """Replace text in a file with new text.
    
    Args:
        file: Input file path to perform replacements on
        what: Text to search for and replace
        to: Text to replace matches with
        output: Optional output file path. If not provided, will modify input file in-place
        
    Returns:
        Path to the modified file (either input file or output file)
        
    Example:
        replace --file=config.txt --what="DEBUG" --to="INFO"
    """
    print(f"replacing {what} to {to} in {file}")
    where = output
    return replace_in_file(file, what, to, where)

@app.command("replace_with_dictionary")
def replace_dict(
    file: Path = typer.Option(..., exists=True, help="rename file with substitution"),
    dictionary: Path = typer.Option(..., exists=True, help="dictionary to load from"),
    output: Optional[Path] = typer.Option(None, help="optional output, will rewrite --file is not output provided"),
    verbose: bool = typer.Option(False, help="if we should output more to console")
) -> Path:
    """Replace multiple text patterns in a file using a JSON dictionary.
    
    Args:
        file: Input file path to perform replacements on
        dictionary: Path to JSON file containing old->new text mappings
        output: Optional output file path. If not provided, will modify input file in-place
        verbose: If True, prints detailed replacement information
        
    Returns:
        Path to the modified file (either input file or output file)
        
    Example:
        replace_with_dictionary --file=config.txt --dictionary=replacements.json --verbose
    """
    print(f"replacing from {dictionary} in {file}")
    # reading the data from the file
    with dictionary.open("r+") as f:
        data = f.read()
    js: dict = json.loads(data)
    print(f"substitutions dictionary is:")
    for k, v in js.items():
        print(f"REPLACE {k} WITH {v}\n")
    where = output
    return replace_from_dict_in_file(file, js, where, verbose)

if __name__ == '__main__':
    app()